import abc
import asyncio
import dataclasses
import typing

import yab.error
import yab.protected
import yab.task
import yab.data
import yab.queue


MessageQueue = yab.queue.Queue[yab.data.Message]


class Send(abc.ABC):
    @abc.abstractmethod
    async def __call__(
        self, content: yab.data.Content, channel: yab.data.Channel
    ) -> None:
        raise NotImplementedError


class Join(abc.ABC):
    @abc.abstractmethod
    async def __call__(self, channel: yab.data.Channel) -> None:
        raise NotImplementedError


@dataclasses.dataclass
class Can:
    join: typing.Optional[Join] = None
    send: typing.Optional[Send] = None


@dataclasses.dataclass
class State:
    ready: bool = False
    can: Can = dataclasses.field(default_factory=Can)


ProtectedState = yab.protected.Protected[State]


class Backend(abc.ABC):
    @abc.abstractmethod
    async def main(self, protected_state: ProtectedState, queue: MessageQueue) -> None:
        raise NotImplementedError


_Backend = typing.TypeVar("_Backend", bound=Backend)


@dataclasses.dataclass(frozen=True)
class RunningBackend(typing.Generic[_Backend]):
    backend: _Backend
    state: ProtectedState
    queue: MessageQueue
    task: asyncio.Task

    async def stop(self) -> None:
        await yab.task.cancel(self.task)


async def run(backend: _Backend) -> RunningBackend[_Backend]:
    async def main() -> None:
        try:
            await backend.main(protected_state, queue)
        finally:
            async with protected_state.lock_for_write() as state:
                state.can = Can()
                state.ready = False
                await queue.clear()

    protected_state = yab.protected.Protected(State())
    queue = MessageQueue()
    return RunningBackend(
        backend=backend,
        state=protected_state,
        queue=queue,
        task=yab.task.create(main(), name=f"main for f{backend} backend"),
    )
