import typing

import aioimport

import yab.backend
import yab.data

if typing.TYPE_CHECKING:
    import twitchio.dataclasses
    import twitchio.ext.commands


class TwitchIOContext(typing.NamedTuple):
    bot: "twitchio.ext.commands.Bot"
    message: "twitchio.dataclasses.Message"


class Twitch(yab.backend.Backend):
    nick: str
    token: str

    def __init__(self, *, nick: str, token: str) -> None:
        super().__init__()
        self.nick = nick
        self.token = token

    async def main(
        self,
        protected_state: yab.backend.ProtectedState,
        queue: yab.backend.MessageQueue,
    ) -> None:
        await aioimport.import_module("twitchio.ext.commands")
        import twitchio.ext.commands  # noqa: E402

        bot = twitchio.ext.commands.Bot(
            irc_token=self.token,
            nick=self.nick,
            # Twitch strips whitespace from messages,
            # so using space as prefix "cancels" commands parsing.
            # If you don't "cancel" command parsing every
            # "unknown" command will produce exception.
            prefix=" ",  # TODO: remove if twitchio ever supports no-commands mode
        )

        class Join(yab.backend.Join):
            async def __call__(self, channel: yab.data.Channel) -> None:
                await bot.join_channels((channel,))

        class Send(yab.backend.Send):
            async def __call__(
                self, content: yab.data.Content, channel: yab.data.Channel,
            ) -> None:
                await bot.get_channel(channel).send(content)

        @bot.listen("event_ready")
        async def on_ready() -> None:
            async with protected_state.lock_for_write() as state:
                state.can.join = Join()
                state.can.send = Send()
                state.ready = True

        # TODO: remove if twitchio ever supports no-commands mode
        @bot.listen("event_command_error")
        async def on_command_error(
            _, error: twitchio.ext.commands.errors.CommandError
        ) -> None:
            if isinstance(error, twitchio.ext.commands.errors.CommandNotFound):
                return
            raise error

        # TODO: remove if twitchio ever supports event_command_error:
        bot.event_command_error = on_command_error

        @bot.listen("event_message")
        async def on_message(message: "twitchio.dataclasses.Message") -> None:
            if message.author.name != self.nick:
                await queue.put(
                    yab.data.Message(
                        content=message.content,
                        channel=message.channel.name,
                        env={
                            TwitchIOContext: TwitchIOContext(bot=bot, message=message,)
                        },
                    )
                )

        await bot.start()
