import asyncio
import contextlib
import dataclasses

import typing

import yab.data
import yab.error
import yab.backend
import yab.middleware
import yab.protected
import yab.task


ChatReadyCallback = typing.Callable[["Chat"], typing.Awaitable]
BackendReadyCallback = typing.Callable[["Context"], typing.Awaitable]
MessageCallback = typing.Callable[["Context", yab.data.Message], typing.Awaitable]
_Callback = typing.TypeVar("_Callback", bound=typing.Callable)


def add_callback(callback: _Callback, callbacks: typing.Set[_Callback]) -> None:
    if callback in callbacks:
        raise yab.error.RunnerError("Callback already added")
    callbacks.add(callback)


def remove_callback(callback: _Callback, callbacks: typing.Set[_Callback]) -> None:
    if callback not in callbacks:
        raise yab.error.RunnerError("Callback not added")
    callbacks.remove(callback)


@dataclasses.dataclass(frozen=True)
class ScheduledCallbacks:
    tasks: typing.Tuple[asyncio.Task, ...]

    async def wait(self) -> None:
        if self.tasks:
            await asyncio.wait(self.tasks)


class Callbacks:
    __chat_ready_callbacks: typing.Set[ChatReadyCallback]
    __backend_ready_callbacks: typing.Set[BackendReadyCallback]
    __message_callbacks: typing.Set[MessageCallback]

    def __init__(self) -> None:
        self.__chat_ready_callbacks = set()
        self.__backend_ready_callbacks = set()
        self.__message_callbacks = set()

    def add_chat_ready_callback(self, callback: ChatReadyCallback) -> None:
        add_callback(callback, self.__chat_ready_callbacks)

    def remove_chat_ready_callback(self, callback: ChatReadyCallback) -> None:
        remove_callback(callback, self.__chat_ready_callbacks)

    def schedule_chat_ready_callbacks(self, chat: "Chat") -> ScheduledCallbacks:
        return ScheduledCallbacks(
            tuple(
                yab.task.create(
                    callback(chat), name=f"chat ready callback for {callback}"
                )
                for callback in self.__chat_ready_callbacks
            )
        )

    def add_backend_ready_callback(self, callback: BackendReadyCallback) -> None:
        add_callback(callback, self.__backend_ready_callbacks)

    def remove_backend_ready_callback(self, callback: BackendReadyCallback) -> None:
        remove_callback(callback, self.__backend_ready_callbacks)

    def schedule_backend_ready_callbacks(
        self, context: "Context"
    ) -> ScheduledCallbacks:
        return ScheduledCallbacks(
            tuple(
                yab.task.create(
                    callback(context), name=f"backend ready callback for {callback}"
                )
                for callback in self.__backend_ready_callbacks
            )
        )

    def add_message_callback(self, callback: MessageCallback) -> None:
        add_callback(callback, self.__message_callbacks)

    def remove_message_callback(self, callback: MessageCallback) -> None:
        remove_callback(callback, self.__message_callbacks)

    def schedule_message_callbacks(
        self, context: "Context", message: yab.data.Message
    ) -> ScheduledCallbacks:
        return ScheduledCallbacks(
            tuple(
                yab.task.create(
                    callback(context, message), name=f"message callback for {callback}"
                )
                for callback in self.__message_callbacks
            )
        )


class Middlewares:
    __middlewares: yab.protected.Protected[typing.Set[yab.middleware.Middleware]]

    def __init__(self) -> None:
        self.__middlewares = yab.protected.Protected(set())

    async def install(self, middleware: yab.middleware.Middleware) -> None:
        async with self.__middlewares.lock_for_write() as middlewares:
            if middleware in middlewares:
                raise yab.error.Error("Middleware already installed")
            middlewares.add(middleware)

    async def uninstall(self, middleware: yab.middleware.Middleware) -> None:
        async with self.__middlewares.lock_for_write() as middlewares:
            if middleware not in middlewares:
                raise yab.error.Error("Middleware not installed")
            middlewares.remove(middleware)

    async def uninstall_all(self) -> None:
        await self.__middlewares.replace(set())

    async def apply(
        self, backend: yab.backend.RunningBackend, message: yab.data.Message
    ) -> None:
        async with self.__middlewares.lock_for_read() as middlewares:
            for middleware in middlewares:
                message.env[middleware] = await middleware.apply(backend, message)


@dataclasses.dataclass
class Context:
    chat: "Chat"
    backend: yab.backend.Backend
    can: yab.backend.Can


class Backends:
    __chat: "Chat"
    __callbacks: Callbacks
    __middlewares: Middlewares
    __backends: yab.protected.Protected[
        typing.Dict[yab.backend.Backend, yab.backend.RunningBackend]
    ]

    def __init__(
        self, *, chat: "Chat", callbacks: Callbacks, middlewares: Middlewares,
    ) -> None:
        self.__chat = chat
        self.__callbacks = callbacks
        self.__middlewares = middlewares
        self.__backends = yab.protected.Protected(dict())

    async def join(self, backend: yab.backend.Backend) -> None:
        async with self.__backends.lock_for_write() as backends:
            if backend in backends:
                raise yab.error.Error("Backend already joined")
            backends[backend] = await yab.backend.run(backend)

    async def leave(self, backend: yab.backend.Backend) -> None:
        async with self.__backends.lock_for_write() as backends:
            if backend not in backends:
                raise yab.error.Error("Backend not joined")
            await backends[backend].stop()
            del backends[backend]

    async def leave_all(self) -> None:
        async with self.__backends.lock_for_write() as backends:
            for backend in backends:
                await backends[backend].stop()
            backends.clear()

    @contextlib.asynccontextmanager
    async def __context(
        self, running_backend: yab.backend.RunningBackend
    ) -> typing.AsyncIterator[Context]:
        async with running_backend.state.lock_for_write() as state:
            yield Context(
                chat=self.__chat, backend=running_backend.backend, can=state.can,
            )

    async def __schedule_running_backend_ready_callbacks(
        self, running_backend: yab.backend.RunningBackend
    ) -> None:
        async with self.__context(running_backend) as context:
            await self.__callbacks.schedule_backend_ready_callbacks(context).wait()

    async def __watch_running_backend_ready(
        self, running_backend: yab.backend.RunningBackend
    ) -> None:
        ready: typing.Optional[bool] = None
        async for state in running_backend.state.stream_updates(write=False):
            if state.ready != ready:
                if state.ready:
                    yab.task.create(
                        self.__schedule_running_backend_ready_callbacks(running_backend)
                    )
                ready = state.ready

    async def __schedule_running_backend_message_callbacks(
        self, running_backend: yab.backend.RunningBackend, message: yab.data.Message
    ) -> None:
        async with self.__context(running_backend) as context:
            await self.__callbacks.schedule_message_callbacks(context, message).wait()

    async def __watch_running_backend_messages(
        self, running_backend: yab.backend.RunningBackend
    ) -> None:
        while True:
            message = await running_backend.queue.get()
            await self.__middlewares.apply(running_backend, message)
            yab.task.create(
                self.__schedule_running_backend_message_callbacks(
                    running_backend, message
                )
            )

    async def __watch_running_backend(
        self, running_backend: yab.backend.RunningBackend
    ) -> None:
        async with yab.task.ContextManager(
            self.__watch_running_backend_ready(running_backend)
        ):
            await self.__watch_running_backend_messages(running_backend)

    async def main(self) -> None:
        tasks: typing.Dict[yab.backend.Backend, asyncio.Task] = dict()
        try:
            async for backends in self.__backends.stream_updates(write=False):
                for backend in backends:
                    if backend not in tasks:
                        tasks[backend] = yab.task.create(
                            self.__watch_running_backend(backends[backend])
                        )
                for backend in tasks:
                    if backend not in backends:
                        await yab.task.cancel(tasks[backend])
                        del tasks[backend]
        finally:
            for task in tasks.values():
                await yab.task.cancel(task)
            tasks.clear()


class Chat:
    callbacks: Callbacks
    middlewares: Middlewares
    backends: Backends
    __running: yab.protected.Protected[bool]

    def __init__(self, *, callbacks: Callbacks = Callbacks()) -> None:
        self.callbacks = callbacks
        self.middlewares = Middlewares()
        self.backends = Backends(
            chat=self, callbacks=self.callbacks, middlewares=self.middlewares
        )
        self.__running = yab.protected.Protected(False)

    async def async_main(self) -> None:
        try:
            await self.__running.replace(True)
            async with yab.task.ContextManager(self.backends.main()):
                self.callbacks.schedule_chat_ready_callbacks(self)
                async for running in self.__running.stream_updates(write=False):
                    if not running:
                        break
        finally:
            await self.backends.leave_all()
            await self.middlewares.uninstall_all()

    def main(self) -> None:
        asyncio.get_event_loop().run_until_complete(self.async_main())

    async def stop(self) -> None:
        await self.__running.replace(False)
