import asyncio
import contextlib
import typing


DO_WRITE = True
_Wrapped = typing.TypeVar("_Wrapped")


class Protected(typing.Generic[_Wrapped]):
    __wrapped: _Wrapped
    __condition: asyncio.Condition

    def __init__(self, wrapped: _Wrapped) -> None:
        self.__wrapped = wrapped
        self.__condition = asyncio.Condition()

    def locked(self) -> bool:
        return self.__condition.locked()

    @contextlib.asynccontextmanager
    async def lock(self, *, write: bool = DO_WRITE) -> typing.AsyncIterator[_Wrapped]:
        async with self.__condition:
            yield self.__wrapped
            if write:
                self.__condition.notify_all()

    @contextlib.asynccontextmanager
    async def lock_for_read(self) -> typing.AsyncIterator[_Wrapped]:
        async with self.lock(write=False) as wrapped:
            yield wrapped

    @contextlib.asynccontextmanager
    async def lock_for_write(self) -> typing.AsyncIterator[_Wrapped]:
        async with self.lock(write=True) as wrapped:
            yield wrapped

    async def replace(self, wrapped: _Wrapped) -> None:
        async with self.lock_for_write():
            self.__wrapped = wrapped

    @contextlib.asynccontextmanager
    async def wait_for_update_and_lock(
        self, *, write: bool = DO_WRITE
    ) -> typing.AsyncIterator[_Wrapped]:
        async with self.lock(write=write):
            await self.__condition.wait()
            yield self.__wrapped

    async def stream_updates(
        self, *, write: bool = DO_WRITE, yield_current: bool = True
    ) -> typing.AsyncIterable[_Wrapped]:
        if yield_current:
            async with self.lock(write=write) as wrapped:
                yield wrapped
        # there must be no await's between read and wait
        # otherwise we may miss change notification
        while True:
            async with self.wait_for_update_and_lock(write=write) as wrapped:
                yield wrapped
