import asyncio
import contextlib
import typing


def done_callback(task: asyncio.Task) -> None:
    try:
        task.result()
    except asyncio.CancelledError:
        pass


def create(
    awaitable: typing.Awaitable, *, name: typing.Optional[str] = None
) -> asyncio.Task:
    task = asyncio.create_task(awaitable, name=name)
    task.add_done_callback(done_callback)
    return task


async def cancel(task: asyncio.Task):
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass


class ContextManager(contextlib.AbstractAsyncContextManager):
    __coro: typing.Awaitable
    __name: typing.Optional[str] = None
    __task: typing.Optional[asyncio.Task] = None

    def __init__(
        self, coro: typing.Awaitable, *, name: typing.Optional[str] = None
    ) -> None:
        self.name = name
        self.__coro = coro

    async def __aenter__(self) -> asyncio.Task:
        self.__task = create(
            self.__coro,
            name=self.name if self.name else f"context task for {self.__coro}",
        )
        return self.__task

    async def __aexit__(self, _, __, ___) -> None:
        assert self.__task is not None
        await cancel(self.__task)
